//
//  Gamepad.swift
//  Quack
//
//  Created by Marco Fattorel on 20/01/2017.
//  Copyright © 2017 Marco Fattorel. All rights reserved.
//

import Foundation
import CoreFoundation
import IOKit
import IOKit.hid
import simd

class Gamepad: Equatable {
	enum PlayerID: Int {
		case Player1 = 0
		case Player2
		case Player3
		case Player4
		case Player5
		case Player6
		case Player7
		case Player8
		case PlayerCount
	}
	
	static private let hidManager: IOHIDManager = IOHIDManagerCreate(kCFAllocatorDefault, IOOptionBits(kIOHIDOptionsTypeNone))
	static var controllers = [PlayerID: Gamepad]()
	
	static func initialize() {
		if IOHIDManagerOpen(hidManager, IOOptionBits(kIOHIDOptionsTypeNone)) != kIOReturnSuccess {
			print("Error initializing GameController")
			return;
		}
		let hidContext = unsafeBitCast(self, to: UnsafeMutableRawPointer.self)
		
		IOHIDManagerRegisterDeviceMatchingCallback(hidManager, controllerConnected, hidContext)
		let criterion = NSMutableArray()
		criterion.add(NSDictionary(objects: [kHIDPage_GenericDesktop, kHIDUsage_GD_GamePad], forKeys: [kIOHIDDeviceUsagePageKey as NSCopying, kIOHIDDeviceUsageKey as NSCopying]))
		criterion.add(NSDictionary(objects: [kHIDPage_GenericDesktop, kHIDUsage_GD_MultiAxisController], forKeys: [kIOHIDDeviceUsagePageKey as NSCopying, kIOHIDDeviceUsageKey as NSCopying]))
		
		IOHIDManagerSetDeviceMatchingMultiple(hidManager, criterion)
		
		IOHIDManagerScheduleWithRunLoop(hidManager, CFRunLoopGetMain(), CFRunLoopMode.defaultMode.rawValue)
		
		
	}
	
	private let device: IOHIDDevice
	fileprivate let playerID: PlayerID
	
	static let deadZonePercent: Float = 0.2
	
	fileprivate var _lThumbXUsageID: CFIndex = 0
	fileprivate var _lThumbYUsageID: CFIndex = 0
	fileprivate var _rThumbXUsageID: CFIndex = 0
	fileprivate var _rThumbYUsageID: CFIndex = 0
	fileprivate var _lTriggerUsageID: CFIndex = 0
	fileprivate var _rTriggerUsageID: CFIndex = 0
	
	fileprivate var _usesHatSwitch: Bool = false
	fileprivate var _dpadLUsageID: CFIndex = 0
	fileprivate var _dpadRUsageID: CFIndex = 0
	fileprivate var _dpadDUsageID: CFIndex = 0
	fileprivate var _dpadUUsageID: CFIndex = 0
	
	fileprivate var _buttonPauseUsageID: CFIndex = 0
	fileprivate var _buttonAUsageID: CFIndex = 0
	fileprivate var _buttonBUsageID: CFIndex = 0
	fileprivate var _buttonXUsageID: CFIndex = 0
	fileprivate var _buttonYUsageID: CFIndex = 0
	fileprivate var _lShoulderUsageID: CFIndex = 0
	fileprivate var _rShoulderUsageID: CFIndex = 0
	struct Snapshot {
		// Extended gamepad data
		// Axes in the range [-1.0, 1.0]
		var dpadX = Double()
		var dpadY = Double()
		
		// Buttons in the range [0.0, 1.0]
		var buttonA = Double()
		var buttonB = Double()
		var buttonX = Double()
		var buttonY = Double()
		var leftShoulder = Double()
		var rightShoulder = Double()
		
		// Axes in the range [-1.0, 1.0]
		var leftThumbstickX = Double()
		var leftThumbstickY = Double()
		var rightThumbstickX = Double()
		var rightThumbstickY = Double()
		
		// Buttons in the range [0.0, 1.0]
		var leftTrigger = Double()
		var rightTrigger = Double()
	}
	var snapshot = Snapshot()
	
	init(with device: IOHIDDevice, player id: PlayerID) {
		self.device = device
		self.playerID = id
		print(IOHIDDeviceGetProperty(device, kIOHIDManufacturerKey as CFString))
		print(IOHIDDeviceGetProperty(device, kIOHIDProductKey as CFString))
	}
	static func ==(lhs: Gamepad, rhs: Gamepad) -> Bool {
		let idl = UnsafeMutablePointer<UInt64>.allocate(capacity: 1)
		IORegistryEntryGetRegistryEntryID(IOHIDDeviceGetService(lhs.device), idl)
		
		let idr = UnsafeMutablePointer<UInt64>.allocate(capacity: 1)
		IORegistryEntryGetRegistryEntryID(IOHIDDeviceGetService(rhs.device), idr)
		
		return idl.pointee == idr.pointee
	}
}
fileprivate func controllerConnected(context: UnsafeMutableRawPointer?, result: IOReturn, sender: UnsafeMutableRawPointer?, device: IOHIDDevice) -> Void {
	guard result == kIOReturnSuccess else {
		return
	}
	
	var player: Gamepad.PlayerID = .PlayerCount
	for k in 0..<Gamepad.PlayerID.PlayerCount.rawValue {
		if Gamepad.controllers.index(forKey: Gamepad.PlayerID(rawValue: k)!) == nil {
			player = Gamepad.PlayerID(rawValue: k)!
			break
		}
	}
	guard player != .PlayerCount else {
		print("too many gamepad")
		return
	}
	let controller = Gamepad(with: device, player: player)
	
	let vid = IOHIDDeviceGetProperty(device, kIOHIDVendorIDKey as CFString) as! UInt
	let pid = IOHIDDeviceGetProperty(device, kIOHIDProductIDKey as CFString) as! UInt
	
	var axisMin: CFIndex = 0
	var axisMax: CFIndex = 256
	
	if vid == 0x054C { // Sony
		if pid == 0x5C4 { // DualShock 4
			print("Sony Dualshock 4 detected.")
			
			controller._lThumbXUsageID = kHIDUsage_GD_X
			controller._lThumbYUsageID = kHIDUsage_GD_Y
			controller._rThumbXUsageID = kHIDUsage_GD_Z
			controller._rThumbYUsageID = kHIDUsage_GD_Rz
			controller._lTriggerUsageID = kHIDUsage_GD_Rx
			controller._rTriggerUsageID = kHIDUsage_GD_Ry
			
			controller._usesHatSwitch = true
			
			controller._buttonPauseUsageID = 0x0A
			controller._buttonAUsageID = 0x02
			controller._buttonBUsageID = 0x03
			controller._buttonXUsageID = 0x01
			controller._buttonYUsageID = 0x04
			controller._lShoulderUsageID = 0x05
			controller._rShoulderUsageID = 0x06
		} else { // MOVE DO NOT WANT
			return
		}
	} else if vid == 0x045E { // Microsoft
		if pid == 0x028E || pid == 0x028F { // 360 wired/wireless
			print("Microsoft Xbox 360 controller detected.")
			
			axisMin = -(1<<15)
			axisMax =  (1<<15)
			
			controller._lThumbXUsageID = kHIDUsage_GD_X
			controller._lThumbYUsageID = kHIDUsage_GD_Y
			controller._rThumbXUsageID = kHIDUsage_GD_Rx
			controller._rThumbYUsageID = kHIDUsage_GD_Ry
			controller._lTriggerUsageID = kHIDUsage_GD_Z
			controller._rTriggerUsageID = kHIDUsage_GD_Rz
			
			controller._dpadLUsageID = 0x0E
			controller._dpadRUsageID = 0x0F
			controller._dpadDUsageID = 0x0D
			controller._dpadUUsageID = 0x0C
			
			controller._buttonPauseUsageID = 0x09
			controller._buttonAUsageID = 0x01
			controller._buttonBUsageID = 0x02
			controller._buttonXUsageID = 0x03
			controller._buttonYUsageID = 0x04
			controller._lShoulderUsageID = 0x05
			controller._rShoulderUsageID = 0x06
		}
	} else if vid == 0x057E { // Nintendo
		if pid == 0x0306 {
			print("Nintendo Wiimote detected.")
			// TODO can we do anything sensible with this?
		}
	}
	
	setupAxis(device: device, element: getAxis(device: device, axis: controller._lThumbXUsageID), dmin: CFIndex(-1.0),  dmax: CFIndex(1.0), rmin: axisMin, rmax: axisMax, deadZonePercent: Gamepad.deadZonePercent)
	setupAxis(device: device, element: getAxis(device: device, axis: controller._lThumbYUsageID),  dmin: CFIndex(1.0), dmax: CFIndex(-1.0), rmin: axisMin, rmax: axisMax, deadZonePercent: Gamepad.deadZonePercent)
	setupAxis(device: device, element: getAxis(device: device, axis: controller._rThumbXUsageID), dmin: CFIndex(-1.0),  dmax: CFIndex(1.0), rmin: axisMin, rmax: axisMax, deadZonePercent: Gamepad.deadZonePercent)
	setupAxis(device: device, element: getAxis(device: device, axis: controller._rThumbYUsageID),  dmin: CFIndex(1.0), dmax: CFIndex(-1.0), rmin: axisMin, rmax: axisMax, deadZonePercent: Gamepad.deadZonePercent)
	
	setupAxis(device: device, element: getAxis(device: device, axis: controller._lTriggerUsageID), dmin: CFIndex(0.0),  dmax: CFIndex(1.0), rmin: 0, rmax: 256, deadZonePercent: 0.0)
	setupAxis(device: device, element: getAxis(device: device, axis: controller._rTriggerUsageID), dmin: CFIndex(0.0),  dmax: CFIndex(1.0), rmin: 0, rmax: 256, deadZonePercent: 0.0)
	
	IOHIDDeviceRegisterInputValueCallback(device, controllerInput, UnsafeMutableRawPointer(Unmanaged.passUnretained(controller).toOpaque()))
	
	let match = [[kIOHIDElementUsagePageKey:kHIDPage_GenericDesktop],[kIOHIDElementUsagePageKey:kHIDPage_Button]] as CFArray
	IOHIDDeviceSetInputValueMatchingMultiple(device, match)
	
	IOHIDDeviceRegisterRemovalCallback(device, controllerDisconnected, UnsafeMutableRawPointer(Unmanaged.passUnretained(controller).toOpaque()))
	
	
	Gamepad.controllers[player] = controller
	print(player.hashValue, terminator: "")
	print(" connected")
}


fileprivate func setupAxis(device: IOHIDDevice, element:  IOHIDElement, dmin: CFIndex, dmax: CFIndex, rmin: CFIndex, rmax: CFIndex, deadZonePercent: Float) {
	IOHIDElementSetProperty(element, kIOHIDElementCalibrationMinKey as CFString, dmin as CFTypeRef);
	IOHIDElementSetProperty(element, kIOHIDElementCalibrationMaxKey as CFString,dmax as CFTypeRef);
	
	IOHIDElementSetProperty(element, kIOHIDElementCalibrationSaturationMinKey as CFString, rmin as CFTypeRef);
	IOHIDElementSetProperty(element, kIOHIDElementCalibrationSaturationMaxKey as CFString, rmax as CFTypeRef);
	
	if deadZonePercent > 0.0 {
		let mid: CFIndex = (rmin + rmax)/2
		let deadZone = CFIndex(Float((rmax - rmin))*(deadZonePercent/2.0))
		
		IOHIDElementSetProperty(element, kIOHIDElementCalibrationDeadZoneMinKey as CFString, (mid - deadZone) as CFTypeRef);
		IOHIDElementSetProperty(element, kIOHIDElementCalibrationDeadZoneMaxKey as CFString, (mid + deadZone) as CFTypeRef);
	}
}

fileprivate func getAxis(device: IOHIDDevice, axis: CFIndex) -> IOHIDElement
{
	let match = NSDictionary(objects: [kHIDPage_GenericDesktop, axis], forKeys: [kIOHIDElementUsagePageKey as NSCopying, kIOHIDElementUsageKey as NSCopying])
	
	let elements: NSArray = IOHIDDeviceCopyMatchingElements(device, match as CFDictionary, IOOptionBits(kIOHIDOptionsTypeNone))!
	if elements.count != 1 {
		print("Warning. Oops, didn't find exactly one axis?")
	}
	
	return elements[0] as! IOHIDElement
}

fileprivate func controllerDisconnected(context: UnsafeMutableRawPointer?, result: IOReturn, sender: UnsafeMutableRawPointer?) -> Void {
	if result == kIOReturnSuccess {
//		let controller = (context as CFTypeRef) as! Gamepad
		let controller = Unmanaged<Gamepad>.fromOpaque(context!).takeUnretainedValue()
		Gamepad.controllers[controller.playerID] = nil
		
//		let keys = Gamepad.controllers.allKeys(forValue: controller)
//		guard keys.count == 1 else {
//			print("controller added more than one!?")
//			return
//		}
//		Gamepad.controllers[keys.first!] = nil
	}
}

fileprivate func controllerInput(context: UnsafeMutableRawPointer?, result: IOReturn, sender: UnsafeMutableRawPointer?, value: IOHIDValue) -> Void {
	guard result == kIOReturnSuccess else {
		return
	}
	
	let controller = Unmanaged<Gamepad>.fromOpaque(context!).takeUnretainedValue()
//	let controller = (context as CFTypeRef) as! Gamepad
	
	let element = IOHIDValueGetElement(value)
	
	let usagePage = Int(IOHIDElementGetUsagePage(element))
	let usage = CFIndex(IOHIDElementGetUsage(element))
	
	let state = IOHIDValueGetIntegerValue(value)
	let analog = IOHIDValueGetScaledValue(value, IOHIDValueScaleType(kIOHIDValueScaleTypeCalibrated))
	
	if usagePage == kHIDPage_Button {
		if usage == controller._buttonPauseUsageID {
			//			if(state) controller.controllerPausedHandler(controller)
		}
		if usage == controller._buttonAUsageID { controller.snapshot.buttonA = Double(state) }
		if usage == controller._buttonBUsageID { controller.snapshot.buttonB = Double(state) }
		if usage == controller._buttonXUsageID { controller.snapshot.buttonX = Double(state) }
		if usage == controller._buttonYUsageID { controller.snapshot.buttonY = Double(state) }
		if usage == controller._lShoulderUsageID { controller.snapshot.leftShoulder = Double(state) }
		if usage == controller._rShoulderUsageID { controller.snapshot.rightShoulder = Double(state) }
		
		if !(controller._usesHatSwitch) {
			if usage == controller._dpadLUsageID {
				controller.snapshot.dpadX = vector_clamp(controller.snapshot.dpadX - (state != 0 ? 1.0 : -1.0), -1.0, 1.0)
			}
			if usage == controller._dpadRUsageID {
				controller.snapshot.dpadX = vector_clamp(controller.snapshot.dpadX + (state != 0 ? 1.0 : -1.0), -1.0, 1.0)
			}
			if usage == controller._dpadDUsageID {
				controller.snapshot.dpadY = vector_clamp(controller.snapshot.dpadY - (state != 0 ? 1.0 : -1.0), -1.0, 1.0)
			}
			if usage == controller._dpadUUsageID {
				controller.snapshot.dpadY = vector_clamp(controller.snapshot.dpadY + (state != 0 ? 1.0 : -1.0), -1.0, 1.0)
			}
		}
	}
	
	if usagePage == kHIDPage_GenericDesktop {
		if usage == controller._lThumbXUsageID {
			controller.snapshot.leftThumbstickX = analog
		}
		if usage == controller._lThumbYUsageID {
			controller.snapshot.leftThumbstickY = analog
		}
		if usage == controller._rThumbXUsageID {
			controller.snapshot.rightThumbstickX = analog
		}
		if usage == controller._rThumbYUsageID {
			controller.snapshot.rightThumbstickY = analog
		}
		if usage == controller._lTriggerUsageID {
			controller.snapshot.leftTrigger = analog
		}
		if usage == controller._rTriggerUsageID {
			controller.snapshot.rightTrigger = analog
		}
		
		if controller._usesHatSwitch && usage == kHIDUsage_GD_Hatswitch {
			switch state {
			case  0: controller.snapshot.dpadX =  0.0; controller.snapshot.dpadY =  1.0; break;
			case  1: controller.snapshot.dpadX =  1.0; controller.snapshot.dpadY =  1.0; break;
			case  2: controller.snapshot.dpadX =  1.0; controller.snapshot.dpadY =  0.0; break;
			case  3: controller.snapshot.dpadX =  1.0; controller.snapshot.dpadY = -1.0; break;
			case  4: controller.snapshot.dpadX =  0.0; controller.snapshot.dpadY = -1.0; break;
			case  5: controller.snapshot.dpadX = -1.0; controller.snapshot.dpadY = -1.0; break;
			case  6: controller.snapshot.dpadX = -1.0; controller.snapshot.dpadY =  0.0; break;
			case  7: controller.snapshot.dpadX = -1.0; controller.snapshot.dpadY =  1.0; break;
			default: controller.snapshot.dpadX =  0.0; controller.snapshot.dpadY =  0.0; break;
			}
		}
	}
}
